import type { ActionCreatorWithOptionalPayload, AsyncThunk, Draft } from "@reduxjs/toolkit";
import { useEffect, useState } from "react";
import type { KilovoltMessage, SubscriptionHandler } from "@strimertul/kilovolt-client";
import { useAppDispatch, useAppSelector } from "~/store";
import apiReducer, { getUserPoints } from "~/store/api/reducer";
import type {
	APIState,
	LoyaltyPointsEntry,
	LoyaltyStorage,
	RequestStatus,
} from "~/store/api/types";

interface LoadStatus {
	load: RequestStatus;
	save: RequestStatus;
}

export const useKilovoltClient = () => useAppSelector((state) => state.api.client);

export function useLiveKeyString(key: string) {
	const client = useKilovoltClient();
	const [data, setData] = useState<string>(null);

	useEffect(() => {
		const subscriber: SubscriptionHandler = (v) => setData(v);
		void client.getKey(key).then((value) => setData(value));
		void client.subscribeKey(key, subscriber);
		return () => {
			void client.unsubscribeKey(key, subscriber);
		};
	}, [key]);

	return data;
}

export function useLiveKey<T>(key: string): T {
	const data = useLiveKeyString(key);
	return data ? (JSON.parse(data) as T) : null;
}

export function useModule<T>({
	key,
	selector,
	getter,
	setter,
	asyncSetter,
}: {
	key: string;
	selector: (state: Draft<APIState>) => T;
	getter: AsyncThunk<T, void, unknown>;
	setter: AsyncThunk<KilovoltMessage, T, unknown>;
	asyncSetter: ActionCreatorWithOptionalPayload<T, string>;
}): [T, AsyncThunk<KilovoltMessage, T, unknown>, LoadStatus] {
	const client = useKilovoltClient();
	const data = useAppSelector((state) => selector(state.api));
	const loadStatus = useAppSelector((state) => state.api.requestStatus[`load-${key}`]);
	const saveStatus = useAppSelector((state) => state.api.requestStatus[`save-${key}`]);
	const dispatch = useAppDispatch();

	useEffect(() => {
		void dispatch(getter());
		const subscriber: SubscriptionHandler = (newValue) => {
			void dispatch(asyncSetter(JSON.parse(newValue) as T));
		};
		void client.subscribeKey(key, subscriber);
		return () => {
			void client.unsubscribeKey(key, subscriber);
			dispatch(apiReducer.actions.requestKeysRemoved([`save-${key}`, `load-${key}`]));
		};
	}, [key, getter, asyncSetter]);

	return [
		data,
		setter,
		{
			load: loadStatus,
			save: saveStatus,
		},
	];
}

const interval = 5000;
/**
 * Sets a timed reactive status based on a RequestStatus.
 * This allows UI elements to be updated only for some time based on the status of a request.
 * Once the status changes, the status on the output will be set for 5 seconds then reset to null.
 * @param status Status to monitor
 * @returns Reactive status
 */
export function useTimedStatus(status: RequestStatus | null): RequestStatus | null {
	const [localStatus, setlocalStatus] = useState(status);
	const [maxTime, setMaxTime] = useState(0);

	useEffect(() => {
		const timeout = Date.now() - interval;
		setMaxTime(timeout);
		const remaining = status ? status.updated.getTime() - timeout : null;
		if (remaining) {
			setTimeout(() => {
				setlocalStatus(null);
			}, remaining);
		}
		setlocalStatus(status);
	}, [status]);

	return status?.updated.getTime() > maxTime ? localStatus : null;
}

export function useUserPoints(): LoyaltyStorage {
	const prefix = "loyalty/points/";
	const client = useKilovoltClient();
	const data = useAppSelector((state) => state.api.loyalty.users);
	const dispatch = useAppDispatch();

	useEffect(() => {
		void dispatch(getUserPoints());
		const subscriber: SubscriptionHandler = (newValue, key) => {
			const user = key.substring(prefix.length);
			const entry = JSON.parse(newValue) as LoyaltyPointsEntry;
			void dispatch(apiReducer.actions.loyaltyUserPointsChanged({ user, entry }));
		};
		void client.subscribePrefix(prefix, subscriber);
		return () => {
			void client.unsubscribePrefix(prefix, subscriber);
		};
	}, []);

	return data;
}

export default {
	useModule,
	useStatus: useTimedStatus,
	useUserPoints,
};
